@@ -29,7 +29,7 @@ from xpra.gtk_common.gtk_util import (
     choose_file, imagebutton, get_icon_pixbuf,
     )
 from xpra.util import DEFAULT_PORT, csv, repr_ellipsized
-from xpra.os_util import WIN32, OSX, POSIX
+from xpra.os_util import WIN32, OSX, is_X11
 from xpra.make_thread import start_thread
 from xpra.gtk_common.about import about
 from xpra.scripts.main import (
@@ -1096,7 +1096,7 @@ def do_main(argv):
     from xpra.scripts.main import InitExit, InitInfo
     from xpra.platform.gui import init as gui_init, ready as gui_ready
 
-    if POSIX and not OSX:
+    if is_X11():
         from xpra.x11.gtk_x11.gdk_display_source import init_gdk_display_source
         init_gdk_display_source()
 
