@@ -11,7 +11,7 @@ import uuid
 from time import monotonic
 
 from xpra.child_reaper import getChildReaper
-from xpra.os_util import bytestostr, umask_context, POSIX, WIN32
+from xpra.os_util import bytestostr, strtobytes, umask_context, POSIX, WIN32
 from xpra.util import typedict, csv, envint, envbool, engs, net_utf8, u
 from xpra.scripts.config import parse_bool, parse_with_unit
 from xpra.simple_stats import std_unit
@@ -270,7 +270,7 @@ class FileTransferHandler(FileTransferAttributes):
                 #transfer has been cancelled
                 return
             chunk_state[-2] = 0     #this timer has been used
-            if chunk_state[-1]==0:
+            if chunk_state[-1]==chunk_no:
                 filelog.error("Error: chunked file transfer '%s' timed out", chunk_id)
                 self.receive_chunks_in_progress.pop(chunk_id, None)
 
@@ -333,6 +333,7 @@ class FileTransferHandler(FileTransferAttributes):
             osclose(fd)
             progress(-1, "chunk no mismatch")
             return
+        file_data = strtobytes(file_data)
         #update chunk number:
         chunk_state[-1] = chunk
         digest = chunk_state[8]
@@ -349,6 +350,11 @@ class FileTransferHandler(FileTransferAttributes):
             osclose(fd)
             progress(-1, "write error (%s)" % e)
             return
+        filesize = chunk_state[6]
+        if written>filesize:
+            filelog.error("Error: too much data received")
+            progress(-1, "file size mismatch")
+            return
         self.send("ack-file-chunk", chunk_id, True, "", chunk)
         if has_more:
             progress(written)
@@ -458,7 +464,7 @@ class FileTransferHandler(FileTransferAttributes):
             l = len(self.receive_chunks_in_progress)
             if l>=MAX_CONCURRENT_FILES:
                 self.send("ack-file-chunk", chunk_id, False, "too many file transfers in progress: %i" % l, 0)
-                os.close(fd)
+                osclose(fd)
                 return
             digest = hashlib.sha256()
             chunk = 0
@@ -721,7 +727,8 @@ class FileTransferHandler(FileTransferAttributes):
             return False
         if ask:
             return self.send_data_request(action, "file", filename, mimetype, data, filesize, printit, openit, options)
-        self.do_send_file(filename, mimetype, data, filesize, printit, openit, options)
+        send_id = uuid.uuid4().hex
+        self.do_send_file(filename, mimetype, data, filesize, printit, openit, options, send_id)
         return True
 
     def send_data_request(self, action, dtype, url, mimetype="", data="", filesize=0, printit=False, openit=True, options=None):
@@ -924,7 +931,7 @@ class FileTransferHandler(FileTransferAttributes):
             filelog.error("Error: cannot find the file transfer id '%r'", chunk_id)
             return
         if chunk_state[-1]!=chunk:
-            filelog.error("Error: chunk number mismatch (%i vs %i)", chunk_state, chunk)
+            filelog.error("Error: chunk number mismatch (%i vs %i)", chunk_state[-1], chunk)
             self.cancel_sending(chunk_id)
             return
         start_time, data, chunk_size, timer, chunk = chunk_state
