@@ -5,30 +5,32 @@
 
 import os.path
 import socket
-from time import sleep
+from time import sleep, monotonic
 from ctypes import Structure, c_uint8, sizeof
 
+from xpra.common import GROUP
 from xpra.scripts.config import InitException, InitExit, TRUE_OPTIONS
 from xpra.exit_codes import (
     EXIT_SSL_FAILURE, EXIT_SSL_CERTIFICATE_VERIFY_FAILURE,
     EXIT_SERVER_ALREADY_EXISTS, EXIT_SOCKET_CREATION_ERROR,
     )
-from xpra.net.bytestreams import set_socket_timeout, pretty_socket
+from xpra.net.bytestreams import set_socket_timeout, pretty_socket, SOCKET_TIMEOUT
 from xpra.os_util import (
-    getuid, get_username_for_uid, get_groups, get_group_id,
-    path_permission_info, monotonic_time, umask_context, WIN32, OSX, POSIX,
+    getuid, get_username_for_uid, get_groups, get_group_id, osexpand,
+    path_permission_info, umask_context, WIN32, OSX, POSIX,
     parse_encoded_bin_data,
     )
 from xpra.util import (
-    envint, envbool, csv, parse_simple_dict,
-    ellipsizer,
+    envint, envbool, csv, parse_simple_dict, print_nested_dict, std,
+    ellipsizer, noerr,
     DEFAULT_PORT,
     )
 from xpra.make_thread import start_thread
 
+#pylint: disable=import-outside-toplevel
+
 #what timeout value to use on the socket probe attempt:
 WAIT_PROBE_TIMEOUT = envint("XPRA_WAIT_PROBE_TIMEOUT", 6)
-GROUP = os.environ.get("XPRA_GROUP", "xpra")
 PEEK_TIMEOUT = envint("XPRA_PEEK_TIMEOUT", 1)
 PEEK_TIMEOUT_MS = envint("XPRA_PEEK_TIMEOUT_MS", PEEK_TIMEOUT*1000)
 UNIXDOMAIN_PEEK_TIMEOUT_MS = envint("XPRA_UNIX_DOMAIN_PEEK_TIMEOUT_MS", 100)
@@ -121,10 +123,9 @@ def hosts(host_str):
         return ["0.0.0.0", "::"]
     return [host_str]
 
-def add_listen_socket(socktype, sock, info, new_connection_cb, new_udp_connection_cb=None, options=None):
+def add_listen_socket(socktype, sock, info, new_connection_cb, options=None):
     log = get_network_logger()
-    log("add_listen_socket(%s, %s, %s, %s, %s, %s)",
-        socktype, sock, info, new_connection_cb, new_udp_connection_cb, options)
+    log("add_listen_socket%s", (socktype, sock, info, new_connection_cb, options))
     try:
         #ugly that we have different ways of starting sockets,
         #TODO: abstract this into the socket class
@@ -134,19 +135,15 @@ def add_listen_socket(socktype, sock, info, new_connection_cb, new_udp_connectio
             sock.start()
             return None
         sources = []
-        if socktype=="udp":
-            assert new_udp_connection_cb, "UDP sockets cannot be handled here"
-            new_udp_connection_cb(sock)
-        else:
-            from gi.repository import GLib
-            sock.listen(5)
-            def io_in_cb(sock, flags):
-                log("io_in_cb(%s, %s)", sock, flags)
-                return new_connection_cb(socktype, sock)
-            source = GLib.io_add_watch(sock, GLib.PRIORITY_DEFAULT, GLib.IO_IN, io_in_cb)
-            sources.append(source)
+        from gi.repository import GLib
+        sock.listen(5)
+        def io_in_cb(sock, flags):
+            log("io_in_cb(%s, %s)", sock, flags)
+            return new_connection_cb(socktype, sock)
+        source = GLib.io_add_watch(sock, GLib.PRIORITY_DEFAULT, GLib.IO_IN, io_in_cb)
+        sources.append(source)
         upnp_cleanup = []
-        if socktype in ("udp", "tcp", "ws", "wss", "ssh", "ssl"):
+        if socktype in ("tcp", "ws", "wss", "ssh", "ssl"):
             upnp = (options or {}).get("upnp", "no")
             if upnp.lower() in TRUE_OPTIONS:
                 from xpra.net.upnp import upnp_add
@@ -160,7 +157,7 @@ def add_listen_socket(socktype, sock, info, new_connection_cb, new_udp_connectio
                     start_thread(c, "pnp-cleanup-%s" % c, daemon=True)
         return cleanup
     except Exception as e:
-        log("add_listen_socket(%s, %s)", socktype, sock, exc_info=True)
+        log("add_listen_socket%s", (socktype, sock, info, new_connection_cb, options), exc_info=True)
         log.error("Error: failed to listen on %s socket %s:", socktype, info or sock)
         log.error(" %s", e)
         return None
@@ -189,11 +186,11 @@ def accept_connection(socktype, listener, timeout=None, socket_options=None):
 
 def peek_connection(conn, timeout=PEEK_TIMEOUT_MS, size=PEEK_SIZE):
     log = get_network_logger()
-    log("peek_connection(%s, %i)", conn, timeout)
+    log("peek_connection(%s, %i, %i)", conn, timeout, size)
     peek_data = b""
-    start = monotonic_time()
+    start = monotonic()
     elapsed = 0
-    set_socket_timeout(conn, PEEK_TIMEOUT_MS*1000)
+    set_socket_timeout(conn, PEEK_TIMEOUT_MS/1000)
     while elapsed<=timeout:
         try:
             peek_data = conn.peek(size)
@@ -205,7 +202,7 @@ def peek_connection(conn, timeout=PEEK_TIMEOUT_MS, size=PEEK_SIZE):
             log("peek_connection(%s, %i) failed", conn, timeout, exc_info=True)
             break
         sleep(timeout/4000.0)
-        elapsed = int(1000*(monotonic_time()-start))
+        elapsed = int(1000*(monotonic()-start))
         log("peek: elapsed=%s, timeout=%s", elapsed, timeout)
     log("socket %s peek: got %i bytes", conn, len(peek_data))
     return peek_data
@@ -244,7 +241,7 @@ def get_sockopt_tcp_info(sock, TCP_INFO, attributes=POSIX_TCP_INFO):
         tcpinfo = tcpinfo_class.from_buffer_copy(data[:tcpinfo_size])
     except ValueError as e:
         log("getsockopt(SOL_TCP, TCP_INFO, %i)", tcpinfo_size, exc_info=True)
-        log("TCPInfo fields=%s", csv(tcpinfo_class._fields_))
+        log("TCPInfo fields=%s", csv(attributes))
         log.warn("Warning: failed to get TCP_INFO for %s", sock)
         log.warn(" %s", e)
         return {}
@@ -261,7 +258,7 @@ def guess_packet_type(data):
         from xpra.net.header import (
             unpack_header, HEADER_SIZE,
             FLAGS_RENCODE, FLAGS_YAML,
-            LZ4_FLAG, LZO_FLAG, BROTLI_FLAG,
+            LZ4_FLAG, BROTLI_FLAG,
             )
         header = data.ljust(HEADER_SIZE, b"\0")
         _, protocol_flags, compression_level, packet_index, data_size = unpack_header(header)
@@ -273,15 +270,21 @@ def guess_packet_type(data):
             rencode = bool(protocol_flags & FLAGS_RENCODE)
             yaml = bool(protocol_flags & FLAGS_YAML)
             lz4 = bool(protocol_flags & LZ4_FLAG)
-            lzo = bool(protocol_flags & LZO_FLAG)
             brotli = bool(protocol_flags & BROTLI_FLAG)
-            #rencode and yaml are mutually exclusive,
-            #so are the compressors
-            compressors = sum((lz4, lzo, brotli))
-            if not (rencode and yaml) and not compressors>1:
-                #if compression is enabled, the compression level must be set:
-                if not compressors or compression_level>0:
-                    pass #return "xpra"
+            def is_xpra():
+                compressors = sum((lz4, brotli))
+                #only one compressor can be enabled:
+                if compressors>1:
+                    return False
+                if compressors==1 and compression_level<=0:
+                    #if compression is enabled, the compression level must be set:
+                    return False
+                if rencode and yaml:
+                    #rencode and yaml are mutually exclusive:
+                    return False
+                return True
+            if is_xpra():
+                return "xpra"
     if data[:4]==b"SSH-":
         return "ssh"
     if data[0]==0x16:
@@ -296,9 +299,8 @@ def guess_packet_type(data):
     return None
 
 
-def create_sockets(opts, error_cb):
+def create_sockets(opts, error_cb, retry=0):
     bind_tcp = parse_bind_ip(opts.bind_tcp)
-    bind_udp = parse_bind_ip(opts.bind_udp)
     bind_ssl = parse_bind_ip(opts.bind_ssl, 443)
     bind_ssh = parse_bind_ip(opts.bind_ssh, 22)
     bind_ws  = parse_bind_ip(opts.bind_ws, 80)
@@ -306,23 +308,7 @@ def create_sockets(opts, error_cb):
     bind_rfb = parse_bind_ip(opts.bind_rfb, 5900)
     bind_vsock = parse_bind_vsock(opts.bind_vsock)
 
-    sockets = {}
-
     min_port = int(opts.min_port)
-    def add_tcp_socket(socktype, host_str, iport, options):
-        if iport!=0 and iport<min_port:
-            error_cb("invalid %s port number %i (minimum value is %i)" % (socktype, iport, min_port))
-        for host in hosts(host_str):
-            sock = setup_tcp_socket(host, iport, socktype)
-            host, iport = sock[2]
-            sockets[sock] = options
-    def add_udp_socket(socktype, host_str, iport, options):
-        if iport!=0 and iport<min_port:
-            error_cb("invalid %s port number %i (minimum value is %i)" % (socktype, iport, min_port))
-        for host in hosts(host_str):
-            sock = setup_udp_socket(host, iport, socktype)
-            host, iport = sock[2]
-            sockets[sock] = options
     # Initialize the TCP sockets before the display,
     # That way, errors won't make us kill the Xvfb
     # (which may not be ours to kill at that point)
@@ -341,6 +327,8 @@ def create_sockets(opts, error_cb):
             sshlog.error(" %s", e)
             ssh_upgrades = False
     log = get_network_logger()
+    #prepare tcp socket definitions:
+    tcp_defs = []
     for socktype, defs in {
         "tcp"   : bind_tcp,
         "ssl"   : bind_ssl,
@@ -351,10 +339,36 @@ def create_sockets(opts, error_cb):
         }.items():
         log("setting up %s sockets: %s", socktype, csv(defs.items()))
         for (host, iport), options in defs.items():
-            add_tcp_socket(socktype, host, iport, options)
-    log("setting up UDP sockets: %s", csv(bind_udp.items()))
-    for (host, iport), options in bind_udp.items():
-        add_udp_socket("udp", host, iport, options)
+            if iport!=0 and iport<min_port:
+                error_cb("invalid %s port number %i (minimum value is %i)" % (socktype, iport, min_port))
+            for h in hosts(host):
+                tcp_defs.append((socktype, h, iport, options, None))
+
+    sockets = {}
+    for attempt in range(retry+1):
+        if not tcp_defs:
+            break
+        try_list = tuple(tcp_defs)
+        tcp_defs = []
+        for socktype, host, iport, options, _ in try_list:
+            try:
+                sock = setup_tcp_socket(host, iport, socktype)
+            except Exception as e:
+                log("setup_tcp_socket%s attempt=%s", (host, iport, options), attempt)
+                tcp_defs.append((socktype, host, iport, options, e))
+            else:
+                host, iport = sock[2]
+                sockets[sock] = options
+        if tcp_defs:
+            sleep(1)
+    if tcp_defs:
+        #failed to create some sockets:
+        for socktype, host, iport, options, exception in tcp_defs:
+            log.error("Error creating %s socket", socktype)
+            log.error(" on %s:%s", host, iport)
+            log.error(" %s", exception)
+            raise InitException("failed to create %s socket: %s" % (socktype, exception))
+
     log("setting up vsock sockets: %s", csv(bind_vsock.items()))
     for (cid, iport), options in bind_vsock.items():
         sock = setup_vsock_socket(cid, iport)
@@ -414,40 +428,6 @@ def setup_tcp_socket(host, iport, socktype="tcp"):
     log.info("created %s socket '%s:%s'", socktype, host, iport)
     return socktype, tcp_socket, (host, iport), cleanup_tcp_socket
 
-def create_udp_socket(host, iport):
-    if host.find(":")<0:
-        listener = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
-        sockaddr = (host, iport)
-    else:
-        assert socket.has_ipv6, "specified an IPv6 address but this is not supported"
-        res = socket.getaddrinfo(host, iport, socket.AF_INET6, socket.SOCK_DGRAM)
-        listener = socket.socket(socket.AF_INET6, socket.SOCK_DGRAM)
-        sockaddr = res[0][-1]
-    listener.bind(sockaddr)
-    return listener
-
-def setup_udp_socket(host, iport, socktype="udp"):
-    log = get_network_logger()
-    try:
-        udp_socket = create_udp_socket(host, iport)
-    except Exception as e:
-        log("create_udp_socket%s", (host, iport), exc_info=True)
-        raise InitExit(EXIT_SOCKET_CREATION_ERROR,
-                       "failed to setup %s socket on %s:%s %s" % (socktype, host, iport, e)) from None
-    def cleanup_udp_socket():
-        log.info("closing %s socket %s:%s", socktype, host, iport)
-        try:
-            udp_socket.close()
-        except OSError:
-            pass
-    if iport==0:
-        iport = udp_socket.getsockname()[1]
-        log.info("allocated UDP port %i for %s", iport, host)
-    log("%s: %s:%s : %s", socktype, host, iport, socket)
-    log.info("created UDP socket %s:%s", host, iport)
-    return socktype, udp_socket, (host, iport), cleanup_udp_socket
-
-
 def parse_bind_ip(bind_ip, default_port=DEFAULT_PORT):
     ip_sockets = {}
     if bind_ip:
@@ -494,7 +474,7 @@ def setup_vsock_socket(cid, iport):
 def parse_bind_vsock(bind_vsock):
     vsock_sockets = {}
     if bind_vsock:
-        from xpra.scripts.main import parse_vsock
+        from xpra.scripts.parsing import parse_vsock  #pylint: disable=import-outside-toplevel
         for spec in bind_vsock:
             parts = spec.split(",", 1)
             cid, iport = parse_vsock(parts[0])
@@ -540,10 +520,11 @@ def setup_local_sockets(bind, socket_dir, socket_dirs, display_name, clobber,
                                   mmap_group, socket_permissions, username, uid, gid))
     if not bind:
         return {}
+    session_dir = os.environ.get("XPRA_SESSION_DIR")
     if not socket_dir and (not socket_dirs or (len(socket_dirs)==1 and not socket_dirs[0])):
         if WIN32:
             socket_dirs = [""]
-        else:
+        elif not session_dir:
             raise InitExit(EXIT_SOCKET_CREATION_ERROR,
                            "at least one socket directory must be set to use unix domain sockets")
     from xpra.platform.dotxpra import DotXpra, norm_makepath
@@ -566,17 +547,20 @@ def setup_local_sockets(bind, socket_dir, socket_dirs, display_name, clobber,
                 assert display_name is not None
                 for sockpath in dotxpra.norm_socket_paths(display_name):
                     sockpaths[sockpath] = options
+                if session_dir and not WIN32:
+                    session_socket = os.path.join(session_dir, "socket")
+                    sockpaths[session_socket] = options
                 log("sockpaths(%s)=%s (uid=%i, gid=%i)", display_name, sockpaths, uid, gid)
             else:
                 sockpath = dotxpra.osexpand(sockpath)
-                if os.path.isabs(sockpath):
-                    pass
-                elif sockpath.endswith("/") or (os.path.exists(sockpath) and os.path.isdir(sockpath)):
+                if sockpath.endswith("/") or (os.path.exists(sockpath) and os.path.isdir(sockpath)):
                     assert display_name is not None
                     sockpath = os.path.abspath(sockpath)
                     if not os.path.exists(sockpath):
                         os.makedirs(sockpath)
                     sockpath = norm_makepath(sockpath, display_name)
+                elif os.path.isabs(sockpath):
+                    pass
                 else:
                     sockpath = dotxpra.socket_path(sockpath)
                 sockpaths[sockpath] = options
@@ -649,8 +633,8 @@ def setup_local_sockets(bind, socket_dir, socket_dirs, display_name, clobber,
                 def timeout_probe(sockpath):
                     #we need a loop because "DEAD" sockets may return immediately
                     #(ie: when the server is starting up)
-                    start = monotonic_time()
-                    while monotonic_time()-start<WAIT_PROBE_TIMEOUT:
+                    start = monotonic()
+                    while monotonic()-start<WAIT_PROBE_TIMEOUT:
                         state = dotxpra.get_server_state(sockpath, WAIT_PROBE_TIMEOUT)
                         log("timeout_probe() get_server_state(%s)=%s", sockpath, state)
                         if state not in (DotXpra.UNKNOWN, DotXpra.DEAD):
@@ -707,10 +691,9 @@ def setup_local_sockets(bind, socket_dir, socket_dirs, display_name, clobber,
         for sock, cleanup_socket in defs.items():
             try:
                 cleanup_socket()
-            except Exception as e:
-                log.error("Error cleaning up socket %s:", sock)
-                log.error(" %s", e)
-                del e
+            except Exception:
+                log.error("Error cleaning up socket %s:", sock, exc_info=True)
+                log.error(" using %s", cleanup_socket)
         raise
     return defs
 
@@ -748,6 +731,13 @@ def handle_socket_error(sockpath, sperms, e):
         raise InitExit(EXIT_SOCKET_CREATION_ERROR,
                        "failed to create socket %s" % sockpath)
 
+def import_zeroconf():
+    from xpra.net.mdns.zeroconf_publisher import ZeroconfPublishers, get_interface_index
+    return ZeroconfPublishers, get_interface_index
+
+def import_avahi():
+    from xpra.net.mdns.avahi_publisher import AvahiPublishers, get_interface_index
+    return AvahiPublishers, get_interface_index
 
 #warn just once:
 MDNS_WARNING = False
@@ -762,21 +752,31 @@ def mdns_publish(display_name, listen_on, text_dict=None):
         from xpra.net import mdns
         assert mdns
         from xpra.net.mdns import XPRA_MDNS_TYPE, RFB_MDNS_TYPE
-        PREFER_ZEROCONF = envbool("XPRA_PREFER_ZEROCONF", True)
-        if PREFER_ZEROCONF:
-            from xpra.net.mdns.zeroconf_publisher import ZeroconfPublishers as MDNSPublishers, get_interface_index
-        else:
-            from xpra.net.mdns.avahi_publisher import AvahiPublishers as MDNSPublishers, get_interface_index
     except ImportError as e:
-        MDNS_WARNING = True
-        log("mdns import failure", exc_info=True)
-        log.warn("Warning: failed to load the mdns publisher")
+        log("mdns support is not installed: %s", e)
+        return ()
+    PREFER_ZEROCONF = envbool("XPRA_PREFER_ZEROCONF", True)
+    imports = [import_zeroconf, import_avahi]
+    if not PREFER_ZEROCONF:
+        imports = reversed(imports)
+    MDNSPublishers = get_interface_index = None
+    exceptions = []
+    for i in imports:
         try:
-            einfo = str(e)
-        except Exception:
-            einfo = str(type(e))
-        log.warn(" %s", einfo)
-        log.warn(" either install the 'python-avahi' module")
+            MDNSPublishers, get_interface_index = i()
+        except ImportError as e:
+            log("mdns import failure", exc_info=True)
+            exceptions.append(e)
+    if not MDNSPublishers or not get_interface_index:
+        MDNS_WARNING = True
+        log.warn("Warning: failed to load the mdns publishers")
+        for e in exceptions:
+            try:
+                einfo = str(e)
+            except Exception:
+                einfo = str(type(e))
+            log.warn(" %s", einfo)
+        log.warn(" install 'python-avahi', 'python-zeroconf'")
         log.warn(" or use the 'mdns=no' option")
         return ()
     d = dict(text_dict or {})
@@ -828,11 +828,103 @@ def get_ssl_attributes(opts, server_side=True, overrides=None):
         args[attr] = v
     return args
 
-def ssl_wrap_socket(sock, **kwargs):
-    fn = get_ssl_wrap_socket_fn(**kwargs)
-    return fn(sock)
+def find_ssl_cert(filename="ssl-cert.pem"):
+    from xpra.log import Logger
+    ssllog = Logger("ssl")
+    #try to locate the cert file from known locations
+    from xpra.platform.paths import get_ssl_cert_dirs  #pylint: disable=import-outside-toplevel
+    dirs = get_ssl_cert_dirs()
+    ssllog("find_ssl_cert(%s) get_ssl_cert_dirs()=%s", filename, dirs)
+    for d in dirs:
+        p = osexpand(d)
+        if not os.path.exists(p):
+            ssllog("ssl cert dir '%s' does not exist", p)
+            continue
+        f = os.path.join(p, "ssl-cert.pem")
+        if not os.path.exists(f):
+            ssllog("ssl cert '%s' does not exist", f)
+            continue
+        if not os.path.isfile(f):
+            ssllog.warn("Warning: '%s' is not a file", f)
+            continue
+        if not os.access(p, os.R_OK):
+            ssllog.info("SSL certificate file '%s' is not accessible", f)
+            continue
+        ssllog("found ssl cert '%s'", f)
+        return f
+    return None
 
-def get_ssl_wrap_socket_fn(cert=None, key=None, ca_certs=None, ca_data=None,
+def ssl_wrap_socket(sock, **kwargs):
+    context, wrap_kwargs = get_ssl_wrap_socket_context(**kwargs)
+    return do_wrap_socket(sock, context, **wrap_kwargs)
+
+def log_ssl_info(ssl_sock):
+    from xpra.log import Logger
+    ssllog = Logger("ssl")
+    ssllog("server_hostname=%s", ssl_sock.server_hostname)
+    cipher = ssl_sock.cipher()
+    if cipher:
+        ssllog.info(" %s, %s bits", cipher[0], cipher[2])
+    try:
+        cert = ssl_sock.getpeercert()
+    except ValueError:
+        pass
+    else:
+        if cert:
+            ssllog("certificate:")
+            print_nested_dict(ssl_sock.getpeercert(), prefix=" ", print_fn=ssllog)
+
+SSL_VERIFY_EXPIRED = 10
+SSL_VERIFY_WRONG_HOST = 20
+SSL_VERIFY_SELF_SIGNED = 18
+SSL_VERIFY_UNTRUSTED_ROOT = 19
+SSL_VERIFY_IP_MISMATCH = 64
+SSL_VERIFY_HOSTNAME_MISMATCH = 62
+SSL_VERIFY_CODES = {
+    SSL_VERIFY_EXPIRED          : "expired",    #also revoked!
+    SSL_VERIFY_WRONG_HOST       : "wrong host",
+    SSL_VERIFY_SELF_SIGNED      : "self-signed",
+    SSL_VERIFY_UNTRUSTED_ROOT   : "untrusted-root",
+    SSL_VERIFY_IP_MISMATCH      : "ip-mismatch",
+    SSL_VERIFY_HOSTNAME_MISMATCH: "hostname-mismatch",
+    }
+
+class SSLVerifyFailure(InitExit):
+    def __init__(self, status, msg, verify_code, ssl_sock):
+        super().__init__(status, msg)
+        self.verify_code = verify_code
+        self.ssl_sock = ssl_sock
+
+def ssl_handshake(ssl_sock):
+    from xpra.log import Logger
+    ssllog = Logger("ssl")
+    try:
+        ssl_sock.do_handshake(True)
+        ssllog.info("SSL handshake complete, %s", ssl_sock.version())
+        log_ssl_info(ssl_sock)
+    except Exception as e:
+        ssllog("do_handshake", exc_info=True)
+        log_ssl_info(ssl_sock)
+        import ssl
+        SSLEOFError = getattr(ssl, "SSLEOFError", None)
+        if SSLEOFError and isinstance(e, SSLEOFError):
+            return None
+        status = EXIT_SSL_FAILURE
+        SSLCertVerificationError = getattr(ssl, "SSLCertVerificationError", None)
+        if SSLCertVerificationError and isinstance(e, SSLCertVerificationError):
+            verify_code = getattr(e, "verify_code", 0)
+            ssllog("verify_code=%s", SSL_VERIFY_CODES.get(verify_code, verify_code))
+            try:
+                msg = getattr(e, "verify_message") or (e.args[1].split(":", 2)[2])
+            except (ValueError, IndexError):
+                msg = str(e)
+            status = EXIT_SSL_CERTIFICATE_VERIFY_FAILURE
+            ssllog("host failed SSL verification: %s", msg)
+            raise SSLVerifyFailure(status, msg, verify_code, ssl_sock) from None
+        raise InitExit(status, "SSL handshake failed: %s" % str(e)) from None
+    return ssl_sock
+
+def get_ssl_wrap_socket_context(cert=None, key=None, ca_certs=None, ca_data=None,
                         protocol="TLSv1_2",
                         client_verify_mode="optional", server_verify_mode="required", verify_flags="X509_STRICT",
                         check_hostname=False, server_hostname=None,
@@ -846,7 +938,7 @@ def get_ssl_wrap_socket_fn(cert=None, key=None, ca_certs=None, ca_data=None,
         verify_mode = server_verify_mode
     from xpra.log import Logger
     ssllog = Logger("ssl")
-    ssllog("get_ssl_wrap_socket_fn%s", (cert, key, ca_certs, ca_data,
+    ssllog("get_ssl_wrap_socket_context%s", (cert, key, ca_certs, ca_data,
                                     protocol,
                                     client_verify_mode, server_verify_mode, verify_flags,
                                     check_hostname, server_hostname,
@@ -857,6 +949,8 @@ def get_ssl_wrap_socket_fn(cert=None, key=None, ca_certs=None, ca_data=None,
     #ca-certs:
     if ca_certs=="default":
         ca_certs = None
+    elif ca_certs=="auto":
+        ca_certs = find_ssl_cert("ca-cert.pem")
     ssllog(" ca_certs=%s", ca_certs)
     #parse verify-mode:
     ssl_cert_reqs = getattr(ssl, "CERT_%s" % verify_mode.upper(), None)
@@ -909,9 +1003,18 @@ def get_ssl_wrap_socket_fn(cert=None, key=None, ca_certs=None, ca_data=None,
     context.options = ssl_options
     ssllog(" cert=%s, key=%s", cert, key)
     if cert:
+        if cert=="auto":
+            #try to locate the cert file from known locations
+            cert = find_ssl_cert()
+            if not cert:
+                raise InitException("failed to automatically locate an SSL certificate to use")
         SSL_KEY_PASSWORD = os.environ.get("XPRA_SSL_KEY_PASSWORD")
         ssllog("context.load_cert_chain%s", (cert or None, key or None, SSL_KEY_PASSWORD))
-        context.load_cert_chain(certfile=cert or None, keyfile=key or None, password=SSL_KEY_PASSWORD)
+        try:
+            context.load_cert_chain(certfile=cert or None, keyfile=key or None, password=SSL_KEY_PASSWORD)
+        except ssl.SSLError as e:
+            ssllog("load_cert_chain", exc_info=True)
+            raise InitException("SSL error, failed to load certificate chain: %s" % e) from e
     if ssl_cert_reqs!=ssl.CERT_NONE:
         if server_side:
             purpose = ssl.Purpose.CLIENT_AUTH   #@UndefinedVariable
@@ -951,44 +1054,168 @@ def get_ssl_wrap_socket_fn(cert=None, key=None, ca_certs=None, ca_data=None,
                 context.load_verify_locations(cafile=f.name)
     elif check_hostname and not server_side:
         ssllog("cannot check hostname client side with verify mode %s", verify_mode)
+    return context, kwargs
+
+def do_wrap_socket(tcp_socket, context, **kwargs):
     wrap_socket = context.wrap_socket
-    def do_wrap_socket(tcp_socket):
-        assert tcp_socket
-        ssllog("do_wrap_socket(%s)", tcp_socket)
-        if WIN32:
-            #on win32, setting the tcp socket to blocking doesn't work?
-            #we still hit the following errors that we need to retry:
-            from xpra.net import bytestreams
-            bytestreams.CAN_RETRY_EXCEPTIONS = (ssl.SSLWantReadError, ssl.SSLWantWriteError)
-        else:
-            tcp_socket.setblocking(True)
+    assert tcp_socket
+    from xpra.log import Logger
+    ssllog = Logger("ssl")
+    ssllog("do_wrap_socket(%s, %s)", tcp_socket, context)
+    import ssl
+    if WIN32:
+        #on win32, setting the tcp socket to blocking doesn't work?
+        #we still hit the following errors that we need to retry:
+        from xpra.net import bytestreams
+        bytestreams.CAN_RETRY_EXCEPTIONS = (ssl.SSLWantReadError, ssl.SSLWantWriteError)
+    else:
+        tcp_socket.setblocking(True)
+    try:
+        ssl_sock = wrap_socket(tcp_socket, **kwargs)
+    except Exception as e:
+        ssllog.debug("wrap_socket(%s, %s)", tcp_socket, kwargs, exc_info=True)
+        SSLEOFError = getattr(ssl, "SSLEOFError", None)
+        if SSLEOFError and isinstance(e, SSLEOFError):
+            return None
+        raise InitExit(EXIT_SSL_FAILURE, "Cannot wrap socket %s: %s" % (tcp_socket, e)) from None
+    return ssl_sock
+
+
+def ssl_retry(e, ssl_ca_certs):
+    SSL_RETRY = envbool("XPRA_SSL_RETRY", True)
+    if not SSL_RETRY:
+        return None
+    if not isinstance(e, SSLVerifyFailure):
+        return None
+    #we may be able to ask the user if we wants to accept this certificate
+    from xpra.log import Logger
+    ssllog = Logger("ssl")
+    verify_code = e.verify_code
+    ssl_sock = e.ssl_sock
+    msg = str(e)
+    del e
+    addr = ssl_sock.getpeername()[:2]
+    port = addr[-1]
+    server_hostname = ssl_sock.server_hostname
+    ssllog("ssl_retry: peername=%s, server_hostname=%s", addr, server_hostname)
+    if verify_code not in (
+        SSL_VERIFY_SELF_SIGNED, SSL_VERIFY_WRONG_HOST,
+        SSL_VERIFY_IP_MISMATCH, SSL_VERIFY_HOSTNAME_MISMATCH,
+        ):
+        ssllog("ssl_retry: %s not handled here", SSL_VERIFY_CODES.get(verify_code, verify_code))
+        return None
+    if not server_hostname:
+        ssllog("ssl_retry: not server hostname")
+        return None
+    ssllog("ssl_retry: server_hostname=%s, ssl verify_code=%s (%i)",
+           server_hostname, SSL_VERIFY_CODES.get(verify_code, verify_code), verify_code)
+    from xpra.platform.paths import get_ssl_hosts_config_dirs
+    from xpra.scripts.pinentry_wrapper import confirm
+    host_dirname = std(server_hostname, extras="-.:#_")+"_%i" % port
+    #self-signed cert:
+    if verify_code==SSL_VERIFY_SELF_SIGNED:
+        if ssl_ca_certs not in ("", "default"):
+            ssllog("self-signed cert does not match %r", ssl_ca_certs)
+            return None
+        #perhaps we already have the certificate for this hostname
+        dirs = get_ssl_hosts_config_dirs()
+        host_dirs = [os.path.join(osexpand(d), host_dirname) for d in dirs]
+        cert_filename = "cert.pem"
+        ssllog("looking for %s in %s", cert_filename, host_dirs)
+        for d in host_dirs:
+            f = os.path.join(d, cert_filename)
+            if os.path.exists(f):
+                ssllog("found certificate for %s: %s", server_hostname, f)
+                ssllog("retrying")
+                return {"ssl_ca_certs" : f}
+        #download the certificate data
+        import ssl
         try:
-            ssl_sock = wrap_socket(tcp_socket, **kwargs)
-        except Exception as e:
-            ssllog.debug("wrap_socket(%s, %s)", tcp_socket, kwargs, exc_info=True)
-            SSLEOFError = getattr(ssl, "SSLEOFError", None)
-            if SSLEOFError and isinstance(e, SSLEOFError):
-                return None
-            raise InitExit(EXIT_SSL_FAILURE, "Cannot wrap socket %s: %s" % (tcp_socket, e))
-        if not server_side:
+            cert_data = ssl.get_server_certificate(addr)
+        except ssl.SSLError:
+            cert_data = None
+        if not cert_data:
+            ssllog("failed to get server certificate from %s", addr)
+            return None
+        ssllog("downloaded ssl cert data for %s: %s", addr, ellipsizer(cert_data))
+        #ask the user if he wants to accept this certificate:
+        title = "SSL Certificate Verification Failure"
+        prompt = "Do you want to accept this certificate?"
+        r = confirm((msg, ), title, prompt)
+        ssllog("run_pinentry_confirm(..) returned %r", r)
+        if not r:
+            return None
+        #if there is an existing host config dir, try to use it:
+        for d in [x for x in host_dirs if os.path.exists(x)]:
             try:
-                ssl_sock.do_handshake(True)
-            except Exception as e:
-                ssllog.debug("do_handshake", exc_info=True)
-                SSLEOFError = getattr(ssl, "SSLEOFError", None)
-                if SSLEOFError and isinstance(e, SSLEOFError):
-                    return None
-                status = EXIT_SSL_FAILURE
-                SSLCertVerificationError = getattr(ssl, "SSLCertVerificationError", None)
-                if SSLCertVerificationError and isinstance(e, SSLCertVerificationError):
-                    try:
-                        msg = e.args[1].split(":", 2)[2]
-                    except (ValueError, IndexError):
-                        msg = str(e)
-                    status = EXIT_SSL_CERTIFICATE_VERIFY_FAILURE
-                    #ssllog.warn("host failed SSL verification: %s", msg)
-                else:
-                    msg = str(e)
-                raise InitExit(status, "SSL handshake failed: %s" % msg)
-        return ssl_sock
-    return do_wrap_socket
+                filename = os.path.join(d, cert_filename)
+                with open(filename, "wb") as f:
+                    f.write(cert_data.encode("latin1"))
+                ssllog.info("saved SSL certificate to %s", filename)
+                return {"ssl_ca_certs" : filename}
+            except OSError:
+                ssllog("failed to save cert data to %r", filename, exc_info=True)
+        #try to create a host config dir:
+        for d in host_dirs:
+            folders = os.path.normpath(d).split(os.sep)
+            #we have to be careful and create the 'ssl' dir with 0o700 permissions
+            #but any directory above that can use 0o755
+            try:
+                ssl_dir_index = len(folders)-1
+                while ssl_dir_index>0 and folders[ssl_dir_index]!="ssl":
+                    ssl_dir_index -= 1
+                if ssl_dir_index>1:
+                    parent = os.path.join(*folders[:ssl_dir_index-1])
+                    ssl_dir = os.path.join(*folders[:ssl_dir_index])
+                    os.makedirs(parent, exist_ok=True)
+                    os.makedirs(ssl_dir, mode=0o700, exist_ok=True)
+                os.makedirs(d, mode=0o700)
+                filename = os.path.join(d, cert_filename)
+                with open(filename, "wb") as f:
+                    f.write(cert_data.encode("latin1"))
+                ssllog.info("saved SSL certificate to %s", filename)
+                return {"ssl_ca_certs" : filename}
+            except OSError:
+                ssllog("failed to save cert data to %r", d, exc_info=True)
+        ssllog.warn("Warning: failed to save certificate data")
+        return None
+    if verify_code in (SSL_VERIFY_WRONG_HOST, SSL_VERIFY_IP_MISMATCH, SSL_VERIFY_HOSTNAME_MISMATCH):
+        #ask the user if he wants to skip verifying the host
+        title = "SSL Certificate Verification Failure"
+        prompt = "Do you want to connect anyway?"
+        r = confirm((msg,), title, prompt)
+        ssllog("run_pinentry_confirm(..) returned %r", r)
+        if r:
+            ssllog.info("retrying without checking the hostname")
+            return {"ssl_check_hostname" : False}
+    return None
+
+
+def socket_connect(host, port, timeout=SOCKET_TIMEOUT):
+    socktype = socket.SOCK_STREAM
+    family = 0  #0 means any
+    try:
+        addrinfo = socket.getaddrinfo(host, port, family, socktype)
+    except Exception as e:
+        raise InitException("cannot get %s address for %s: %s" % ({
+            socket.AF_INET6 : "IPv6",
+            socket.AF_INET  : "IPv4",
+            0               : "any",
+            }.get(family, ""), (host, port), e)) from None
+    log = get_network_logger()
+    log("socket_connect%s addrinfo=%s", (host, port), addrinfo)
+    #try each one:
+    for addr in addrinfo:
+        sockaddr = addr[-1]
+        family = addr[0]
+        sock = socket.socket(family, socktype)
+        sock.settimeout(timeout)
+        try:
+            log("socket.connect(%s)", sockaddr)
+            sock.connect(sockaddr)
+            sock.settimeout(None)
+            return sock
+        except Exception as e:
+            log("failed to connect using %s%s for %s", sock.connect, sockaddr, addr, exc_info=True)
+            noerr(sock.close)
+    return None
