@@ -51,12 +51,24 @@ if frozen:
     for d in (libdir, edir):
         if not os.path.exists(d) or not os.path.isdir(d):
             continue
-        if d not in sys.path:
-            sys.path.insert(0, d)
-        if d not in PATH:
-            PATH.insert(0, d)
+        try:
+            sys.path.remove(d)
+        except ValueError:
+            pass
+        sys.path.insert(0, d)
+        try:
+            PATH.remove(d)
+        except ValueError:
+            pass
+        PATH.insert(0, d)
     os.environ['GI_TYPELIB_PATH'] = os.path.join(libdir, "girepository-1.0")
     os.environ["PATH"] = os.pathsep.join(PATH)
+    if not os.environ.get("GTK_THEME") and not os.environ.get("GTK_DEBUG"):
+        for theme in ("Windows-10", "win32"):
+            tdir = os.path.join(edir, "share", "themes", theme)
+            if os.path.exists(tdir):
+                os.environ["GTK_THEME"] = theme
+                break
 
 if REDIRECT_OUTPUT:
     FIX_UNICODE_OUT = False
@@ -70,7 +82,7 @@ def is_wine():
         import winreg   #@UnresolvedImport @Reimport
         hKey = winreg.OpenKey(win32con.HKEY_LOCAL_MACHINE, r"Software\\Wine")   #@UndefinedVariable
         return hKey is not None
-    except:
+    except Exception:
         #no wine key, assume not present and wait for input
         pass
     return False
@@ -82,8 +94,8 @@ def get_csidl_folder(csidl):
         shell32 = WinDLL("shell32", use_last_error=True)
         SHGetFolderPath = shell32.SHGetFolderPathW
         SHGetFolderPath(0, csidl, None, 0, buf)
-        return buf.value
-    except:
+        return os.path.normpath(buf.value)
+    except Exception:
         return None
 
 def get_common_startmenu_dir():
@@ -105,7 +117,7 @@ def set_prgname(name):
     prg_name = name
     try:
         SetConsoleTitleA(name)
-    except:
+    except Exception:
         pass
 
 def not_a_console(handle):
@@ -274,12 +286,12 @@ def get_console_position(handle):
         cpos = csbi.dwCursorPosition
         #wait for input if this is a brand new console:
         return cpos.X and cpos.Y
-    except:
+    except Exception:
         e = sys.exc_info()[1]
         code = -1
         try:
             code = e.winerror
-        except:
+        except AttributeError:
             pass
         if code==errno.ENXIO:
             #ignore "no device" errors silently
@@ -289,7 +301,7 @@ def get_console_position(handle):
         else:
             try:
                 print("error accessing console %s: %s" % (errno.errorcode.get(e.errno, e.errno), e))
-            except:
+            except Exception:
                 print("error accessing console: %s" % e)
         return -1, -1
 
