@@ -231,6 +231,7 @@ def guess_xpra_display(socket_dir, socket_dirs):
 
 def show_encoding_help(opts):
     #avoid errors and warnings:
+    opts.pidfile = None
     opts.encoding = ""
     opts.clipboard = False
     opts.notifications = False
@@ -506,7 +507,7 @@ def apply_config(opts, mode):
     options = load_options()
     if not options:
         return mode
-    if mode=="upgrade":
+    if mode.find("upgrade")>=0:
         #unspecified upgrade, try to find the original mode used:
         mode = options.pop("mode", mode)
     upgrade_config = dict_to_validated_config(options)
@@ -966,6 +967,10 @@ def _do_run_server(script_file, cmdline,
     if upgrading:
         #if we had saved the start / start-desktop config, reload it:
         mode = apply_config(opts, mode)
+        if mode.startswith("upgrade-"):
+            mode = mode[len("upgrade-"):]
+        if mode.startswith("start-"):
+            mode = mode[len("start-"):]
         upgrading_desktop = mode=="desktop"
         upgrading_seamless = not upgrading_desktop
 
@@ -1302,13 +1307,13 @@ def _do_run_server(script_file, cmdline,
             #all unix domain sockets:
             ud_paths = [sockpath for stype, _, sockpath, _ in local_sockets if stype=="unix-domain"]
             if ud_paths:
+                os.environ["XPRA_SERVER_SOCKET"] = ud_paths[0]
                 #choose one so our xdg-open override script can use to talk back to us:
                 if opts.forward_xdg_open:
                     for x in ("/usr/libexec/xpra", "/usr/lib/xpra"):
                         xdg_override = os.path.join(x, "xdg-open")
                         if os.path.exists(xdg_override):
                             os.environ["PATH"] = x+os.pathsep+os.environ.get("PATH", "")
-                            os.environ["XPRA_SERVER_SOCKET"] = ud_paths[0]
                             break
             else:
                 log.warn("Warning: no local server sockets,")
@@ -1447,7 +1452,7 @@ def attach_client(options, defaults):
             continue
         if c!=d:
             if OPTION_TYPES.get(x)==list:
-                v = ",".join(str(i) for i in x)
+                v = ",".join(str(i) for i in c)
             else:
                 v = str(c)
             cmd.append("--%s=%s" % (x, v))
