@@ -670,7 +670,8 @@ class WindowModel(BaseWindowModel):
             geomlog("update_size_constraints%s less restrictive, no need to recalculate", (minw, minh, maxw, maxh))
             return
         geomlog("update_size_constraints%s recalculating client geometry", (minw, minh, maxw, maxh))
-        self._update_client_geometry()
+        if self._setup_done and self.get_property("owner"):
+            self._update_client_geometry()
 
     #########################################
     # X11 properties synced to Python objects
@@ -745,7 +746,7 @@ class WindowModel(BaseWindowModel):
         if 0<cmaxh<hmaxh:
             hmaxh = cmaxh
         #if the values mean something, expose them:
-        if hminw>0 or hminw>0:
+        if hminw>0 or hminh>0:
             hints["minimum-size"] = hminw, hminh
         if hmaxw<MAX_WINDOW_SIZE or hmaxh<MAX_WINDOW_SIZE:
             hints["maximum-size"] = hmaxw, hmaxh
@@ -759,7 +760,7 @@ class WindowModel(BaseWindowModel):
         # reduces the chance for us to get caught in loops:
         if self._updateprop("size-hints", hints):
             metalog("updated: size-hints=%s", hints)
-            if self._setup_done:
+            if self._setup_done and self.get_property("owner"):
                 self._update_client_geometry()
 
 
